function varargout = ntfmcmc(x, K, N, varargin)
% Probabilistic non-negative tensor factorization using Hamiltonian Markov
% chain Monte Carlo.
%
% Usage:
%    ntfmcmc(x, K, N, [options])
%
%    x : Data tensor (missing values set to NaN)
%    K : Number of factors
%    N : Number of iterations
%
% Options:
%     method    : 'hmcmc' or 'gibbs'
%     tau       : Number of leapfrog steps (default 20)
%     epsilon   : Step size (default 1e-3)
%
%   Parameters
%     u         : Latent factors (stacked)
%     v       : Noise variance
%
%   Prior for noise variance
%      alpha, beta : Prior for variance
%
%   Prior for factors
%     mu        : "Mean" of rect. Gauss
%     vv         : "Variance" of rect. Gauss.
%
%   Hyperprior for factors
%     a, b     
%     mumu, vmu
%
% Copyright Mikkel N. Schmidt

M = size(x);
I = ndims(x);
missval = any(isnan(x(:)));
J = sum(~isnan(x(:)));
if missval
    Z = ~isnan(x); 
    x(~Z) = 0;
else
    Z = 1;
end
xx = x(:)'*x(:);
ek = reshape(eyet(K,3),K,[]);

% Parse input arguments
opts = mgetopt(varargin);

% Options
method = mgetopt(opts, 'method', 'hmcmc', 'instrset', {'hmcmc', 'gibbs'});
tau = mgetopt(opts, 'tau', 20);
epsilon = mgetopt(opts, 'epsilon', 1e-3);

% Parameters
u = mgetopt(opts, 'u', rand(sum(M),K));
mu = mgetopt(opts, 'mu', zeros(sum(M),K));
v = mgetopt(opts, 'v', ones(sum(M),K));
vv = mgetopt(opts, 'vv', 1);

% Hyper-parameters
a = mgetopt(opts, 'a', 1);
b = mgetopt(opts, 'b', 1);
alpha = mgetopt(opts, 'alpha', 1);
beta = mgetopt(opts, 'beta', 1);
mumu = mgetopt(opts, 'mumu', 0);
vmu = mgetopt(opts, 'vmu', 1);

% Allocate output arrays
if nargout>=1, varargout{1} = zeros(sum(M),K,N); end
if nargout>=2, varargout{2} = zeros(sum(M),K,N); end
if nargout>=3, varargout{3} = zeros(sum(M),K,N); end
if nargout>=4, varargout{4} = zeros(N,1); end
if nargout>=5, varargout{5} = zeros(N,1); end
if nargout>=6, varargout{6} = zeros(N,1); end

i1 = cumsum([1 M(1:end-1)]);
i2 = cumsum(M);

mprogress;
for n = 1:N            
    switch method
        case 'hmcmc'
            % Compute values in first iterations
            if n==1, 
                ut = max(log(u),-10); 
                vt = max(log(v),-10);
                vvt = max(log(vv),-10);
                [gut,gmu,gvt,gvvt,L,chi,u,mu,v,vv] = hmcmc_update(x,missval,Z,xx,ut,mu,vt,vvt,alpha,beta,a,b,mumu,vmu,ek,I,K,M,J,i1,i2); 
            end
            
            % store variables
            gutnew = gut; gmunew = gmu; gvtnew = gvt; gvvtnew = gvvt;
            utnew = ut; munew = mu; vtnew = vt; vvtnew = vvt;

            % Random momentum
            put = randn(size(ut)); pmu = randn(size(mu)); pvt = randn(size(vt)); pvvt = randn(size(vvt));
            
            % Compute Hamiltonian
            H = (sum(put(:).^2)+sum(pmu(:).^2)+sum(pvt(:).^2)+pvvt.^2)/2 + L;
            
            % Make tau leap-frog steps
            for t = 1:tau                
                put = put-epsilon*gutnew/2; pmu = pmu-epsilon*gmunew/2; pvt = pvt-epsilon*gvtnew/2; pvvt = pvvt-epsilon*gvvtnew/2;                
                utnew = utnew+epsilon*put; munew = munew+epsilon*pmu; vtnew = vtnew+epsilon*pvt; vvtnew = vvtnew+epsilon*pvvt;
                [gutnew,gmunew,gvtnew,gvvtnew] = hmcmc_update(x,missval,Z,xx,utnew,munew,vtnew,vvtnew,alpha,beta,a,b,mumu,vmu,ek,I,K,M,J,i1,i2);
                put = put-epsilon*gutnew/2; pmu = pmu-epsilon*gmunew/2; pvt = pvt-epsilon*gvtnew/2; pvvt = pvvt-epsilon*gvvtnew/2;                
            end
            [gutnew,gmunew,gvtnew,gvvtnew,Lnew,chinew,u,mu,v,vv] = hmcmc_update(x,missval,Z,xx,utnew,munew,vtnew,vvtnew,alpha,beta,a,b,mumu,vmu,ek,I,K,M,J,i1,i2);
            Hnew = (sum(put(:).^2)+sum(pmu(:).^2)+sum(pvt(:).^2)+pvvt.^2)/2 + Lnew;
            
            % Accept step?
            dH = Hnew - H;
            if dH<0 || rand<exp(-dH)
                ut = utnew; mu = munew; vt = vtnew; vvt = vvtnew;
                L = Lnew; chi = chinew;
                gut = gutnew; gmu = gmunew; gvt = gvtnew; gvvt = gvvtnew;
            end    
        case 'gibbs'
            % Sample u
            for i = 1:I
                for k = 1:K
                    [l,r] = compute_lr(u,ek,I,K,i1,i2); % precompute this
                    if missval
                        uik = (reshape((l{i}(:,k).^2)',1,[])*reshape(r{i}(k,:).^2*reshape(Z,[],prod(M(i+1:end)))',[],M(i)))';
                    else
                        uik = sum(sum((l{i}(:,k).^2)*(r{i}(k,:).^2))); 
                    end
                    vik = 1./(uik./vv+1./v(i1(i):i2(i),k));
                    xuik = (reshape((l{i}(:,k))',1,[])*reshape(r{i}(k,:)*reshape(x,[],prod(M(i+1:end)))',[],M(i)))';
                    nk = [1:k-1 k+1:K];
                    xhk = reshape(l{floor(I/2)+1}(:,nk)*r{floor(I/2)}(nk,:),M);
                    uuik = (reshape((l{i}(:,k))',1,[])*reshape(r{i}(k,:)*reshape(xhk.*Z,[],prod(M(i+1:end)))',[],M(i)))';
                    mik = vik.*((-xuik+uuik)./vv+mu(i1(i):i2(i),k)./v(i1(i):i2(i),k));
                    u(i1(i):i2(i),k) = randrg(-mik,vik);
                end
            end
            % Sample vv
            [l,r] = compute_lr(u,ek,I,K,i1,i2); % precompute this
            xu = compute_xu(l,r,x,ek,I,K,M,i1,i2);            
            xh = compute_xh(l,r,I,M);
            uu = compute_xu(l,r,xh.*Z,ek,I,K,M,i1,i2);
            chi = xx+sum(sum(uu(i1(1):i2(1),:).*u(i1(1):i2(1),:)))-2*sum(sum(xu(i1(1):i2(1),:).*u(i1(1):i2(1),:)));
            vv = 1./gamrnd(alpha+.5*J, 1./(beta+.5*chi));
            % Sample mu
            vmub = 1./(1./v+1./vmu);
            mmub = vmub.*(u./v+mumu./vmu);
            mu = mmub+randn(sum(M),K).*sqrt(vmub);
            % Sample v
            v = 1./gamrnd(a+.5,1./(b+.5*(u-mu).^2));
            L = chi./2./vv+(J/2+alpha+1).*log(vv)+beta./vv+ ...
                sum(((u(:)-mu(:)).^2+2*b)./(2*v(:))+(mu(:)-mumu).^2/2/vmu+(a+1)*log(v(:)));
    end

    % Store outputs
    if nargout>=1, varargout{1}(:,:,n) = u; end;
    if nargout>=2, varargout{2}(:,:,n) = mu; end;
    if nargout>=3, varargout{3}(:,:,n) = v; end;
    if nargout>=4, varargout{4}(n) = vv; end;
    if nargout>=5, varargout{5}(n) = L; end;
    if nargout>=6, varargout{6}(n) = chi; end;
    
    % Display timer
    mprogress(n/N);    
%     [l,r] = compute_lr(u,ek,I,K,i1,i2); % precompute this
%     xh = compute_xh(l,r,I,M);
%     imagesc(xh);
%     drawnow;
%     disp(L);
end

function [gut,gmu,gvt,gvvt,L,chi,u,mu,v,vv] = hmcmc_update(x,missval,Z,xx,ut,mu,vt,vvt,alpha,beta,a,b,mumu,vmu,ek,I,K,M,J,i1,i2)
% Changed variables
u = exp(ut);
v = exp(vt);
vv = exp(vvt);
% Intermediates
[l,r] = compute_lr(u,ek,I,K,i1,i2);
xu = compute_xu(l,r,x,ek,I,K,M,i1,i2);
if missval
    xh = compute_xh(l,r,I,M);
    uu = compute_xu(l,r,xh.*Z,ek,I,K,M,i1,i2);
else
    uu = compute_uu(u,I,M,K,i1,i2);
end
% Sum of squared error
chi = xx+sum(sum(uu(i1(1):i2(1),:).*u(i1(1):i2(1),:)))-2*sum(sum(xu(i1(1):i2(1),:).*u(i1(1):i2(1),:)));
% Negative log-likelihood
L = chi./2./vv+(J/2+alpha).*vvt+beta./vv+ ...
    sum(((u(:)-mu(:)).^2+2*b)./(2*v(:))+(mu(:)-mumu).^2/2/vmu+a*vt(:)-ut(:));
% Compute gradients
gut = u.*(uu-xu)./vv+u.*(u-mu)./v-1;
gmu = -(u-mu)./v+(mu-mumu)/vmu;
gvt = -(.5*(u-mu).^2+b)./(v)+a;
gvvt = -(.5*chi+beta)/vv+.5*J+alpha;

function [l,r] = compute_lr(u,ek,I,K,i1,i2)
l = cell(I,1);
r = cell(I,1);
l{1} = ones(1,K);
r{I} = ones(K,1);
for i = 1:I-1
    l{i+1} = reshape(reshape(ek'*l{i}',K,[])'*u(i1(i):i2(i),:)',K,[])';
    r{I-i} = reshape(u(i1(I-i+1):i2(I-i+1),:)*reshape(r{I-i+1}'*ek,[],K)',[],K)';
end

function xh = compute_xh(l,r,I,M)
xh = reshape(l{floor(I/2)+1}*r{floor(I/2)},M);

function xu = compute_xu(l,r,x,ek,I,K,M,i1,i2)
xu = zeros(sum(M),K);
for i = 1:I
    if prod(M(1:i-1))<prod(M(i+1:end))
        xu(i1(i):i2(i),:) = (reshape((l{i}*ek)',K,[])*reshape(r{i}*reshape(x,[],prod(M(i+1:end)))',[],M(i)))';
    else
        xu(i1(i):i2(i),:) = reshape(reshape(x,prod(M(1:i-1)),[])'*l{i},M(i),[])*reshape(r{i}'*ek,[],K);
    end
end

function uu = compute_uu(u,I,M,K,i1,i2)
uuu = zeros(K,K,I);
for i=1:I
    uuu(:,:,i) = (u(i1(i):i2(i),:)'*u(i1(i):i2(i),:));
end
uu = ones(sum(M),K);
for j = 1:I
    uu(i1(j):i2(j),:) = u(i1(j):i2(j),:)*prod(uuu(:,:,[1:j-1 j+1:I]),3);
end
